from __future__ import annotations

import ast
from os.path import exists, isdir

from simudator.core.processor import Processor


class CLI:
    """
    Terminal based program for controlling a processor.
    Mostly used for testing and debugging.
    """

    def __init__(self, processor: Processor):
        self.running = True
        self.processor = processor
        self.print_list = []

    def run(self) -> None:
        """
        Runs the CLI.
        Will put the user in an input loop where they can send commands to the program
        """

        print("Welcome to simuDAtor!")

        while self.running:
            # Get user commands here
            user_input = input("> ")

            match user_input.split():
                case ["hi"] | ["hello"]:
                    print("Greeting!")

                case ["c"] | ["clock"]:
                    # shows current clockcycle
                    print("Clock is at: ", str(self.processor.get_clock()))

                case ["p"] | ["print"]:
                    print("\n")
                    # prints all module states, is most prbably hard to read
                    for module in self.processor.get_modules():
                        module.print_module()
                        print("\n")

                case ["p", "br"] | ["print", "breaks"]:
                    self.processor.print_breakpoints()

                case ["p", "br", "lam"] | ["print", "break", "lambdas"]:
                    lambdas = self.processor.get_breakpoint_lambdas()
                    if not lambdas:
                        print("There are no lambdas for this processor.")
                    else:
                        print(lambdas)

                case ["p", *_] | ["print", *_]:
                    print("\n")
                    # print modules with given names
                    for name in user_input.split()[1:]:
                        try:
                            self.processor.get_module(name).print_module()
                            print("\n")
                        except KeyError:
                            print("\nNo module named", name, "\n")

                case ["n"] | ["next"]:
                    # runs next tick in the processor
                    # this could take time for larger simulations,
                    # could be good to add loading icon and stop user input

                    self.processor.do_tick()

                case ["rc"] | ["run_continuously"]:
                    self.processor.run_continuously()

                case ["r"] | ["resets"]:
                    # self.processor.load_cycle(0)
                    self.processor.reset()

                case ["n", _] | ["next", _]:
                    # runs a given number of ticks
                    # this could take time for larger simulations,
                    # could be good to add loading icon and stop user input

                    try:
                        for _ in range(int(user_input.split()[1])):
                            self.processor.do_tick()
                    except ValueError:
                        print("Invalid value")

                case ["l", _] | ["load", _]:
                    # load processor state from file
                    try:
                        self.processor.reset()
                        self.processor.load_state_from_file(user_input.split()[1])
                    except FileNotFoundError:
                        print("No file called ", user_input.split()[1], " was found")

                case ["pl", *_] | ["printlist", *_]:
                    self.print_list_command(user_input.split())

                case ["u", _] | ["undo", _]:
                    # undo many clock cycles
                    number = 1

                    try:
                        number = int(user_input.split()[1])
                    except ValueError:
                        print("Invalid value")
                        continue

                    current_cycle = self.processor.get_clock()
                    try:
                        self.processor.load_cycle(max(current_cycle - number, 0))
                    except IndexError:
                        print("Index out of range")

                case ["u"] | ["undo"]:
                    # undo one clock cycles

                    current_cycle = self.processor.get_clock()
                    try:
                        self.processor.load_cycle(max(current_cycle - 1, 0))
                    except IndexError:
                        print("Index out of range")

                case ["br", "state", _, _, _] | ["break", "state", _, _, _]:
                    module_name = user_input.split()[2]
                    state_name = user_input.split()[3]
                    val = ast.literal_eval(user_input.split()[4])
                    try:
                        self.processor.add_state_breakpoint(
                            module_name, state_name, val
                        )
                    except ValueError as e:
                        print(e)

                case ["br", "mem", _, _, _] | ["break", "memory", _, _, _]:
                    module_name = user_input.split()[2]
                    try:
                        adress = ast.literal_eval(user_input.split()[3])
                        value = ast.literal_eval(user_input.split()[4])
                        self.processor.add_memory_breakpoint(module_name, adress, value)
                    except ValueError as e:
                        print(e)

                case ["br", "lam", _, *_] | ["break", "lambda", _, *_]:
                    lambda_name = user_input.split()[2]
                    kwargs_str = user_input.split()[3:]
                    kwargs = {}
                    for kwarg in kwargs_str:
                        key, value = kwarg.split('=')
                        kwargs[key] = ast.literal_eval(value)
                    try:
                        self.processor.add_lambda_breakpoint(lambda_name, **kwargs)
                    except ValueError as e:
                        print(e)

                case ["rm", "br", _] | ["remove", "break", _]:
                    try:
                        bp_id = int(user_input.split()[2])
                        if not self.processor.remove_breakpoint(bp_id):
                            print(f"No breakpoint with ID {bp_id}")
                    except ValueError:
                        print("Breakpoint ID must be an integer")

                case ["h"] | ["help"]:
                    # prints help message
                    print(
                        "Here is a list of possible commands:\n"
                        "- c, clock : Prints the processors current clock cycle.\n"
                        "- p [name*], print [name*]: Prints the states with "
                        "given name. "
                        "Will print all modules if no names are given.\n"
                        "- p br, print breaks: Prints all breakpoints.\n"
                        "- p br lam, print break lambdas: Prints all available "
                        "lambdas of this processor to use for lambda "
                        "breakpoints. \n"
                        "- pl, printlist [add] [remove]: "
                        "Prints all modules in the print list. "
                        "Add or remove followed by a modules name "
                        "will add or remove them from the print list.\n"
                        "- l, load [path]: "
                        "Loads a processors state from a given file path.\n"
                        "- r, reset : Resets processor to initial state.\n"
                        "- n [x], next [x]: Simulates x ticks. "
                        "If no x is given simulates one tick.\n"
                        "- rc, run_continuously: "
                        "Continuously simulate ticks until a HALT is signalled "
                        "or a breakpoint is reached.\n"
                        "- u [x], undo [x]: Undos the last x clock cycle. "
                        "If no x is given one cycle is undone.\n"
                        "- br, break [module name] [state name] [value]: "
                        "Add a breakpoint for the module with the given name. "
                        "The breakpoint brakes when the specified state of the "
                        "module has the specified value.\n"
                        "- br mem, break memory [module name] [adress] [value]: "
                        "Add a breakpoint for a memory module with the given "
                        "name. The breakpoint breaks when the specified adress "
                        "of the memory has the specified value.\n"
                        "- br lam, break lambda [lambda_name] "
                        "[keyword argument] ...: Add a breakpoint that uses "
                        "the referenced lambda to evaluate when to break. Any "
                        "module names in the keyword arguments will be "
                        "substituted with the actual module. Strings must be "
                        "enclosed with quotation marks. \n"
                        "- rm br [id], remove break [id]: Remove the "
                        "breakpoint with the given ID.\n"
                        "- h, help : Shows this message.\n"
                        "- q, quit : Closes the program. Nothing will be saved."
                    )

                case ["q"] | ["quit"]:
                    # Ends the loop
                    print("Goodbye")
                    self.running = False

                case ["pp"] | ["pretty print"]:
                    # pretty print the modules of the processor
                    self.processor.pretty_print()

                case ["ppv"] | ["pretty print verbose"]:
                    # pretty print the modules of the processor
                    # with all available information
                    self.processor.pretty_print_verbose()

                case ["sf"] | ["save to file"]:
                    file_path = "cpu_save_file"
                    self.processor.save_state_to_file(file_path)

                case ["sf", _] | ["save to file", _]:
                    file_path = user_input.split()[1]
                    self.check_file_path(file_path)

                case _:
                    # message for unknown command
                    print("Unknown command. Write \"help\" for assistance.")

    def check_file_path(self, file_path) -> None:
        """
        Checks the file path user gave as input. If it is a directory
        this funciton will simply return and not write anything. If it
        is a file that already exists it will demand a 'yes' or 'no' from
        user to decide if it should write over the existing file.
        If the given file does not exist it will simply create the file.
        """

        # Given file does not exist, save_state_to_file will create it
        if not exists(file_path):
            self.processor.save_state_to_file(file_path)

        else:
            # We dont touch directories
            if isdir(file_path):
                print("Given file is a directory")
                return

            # Should we overwrite file?
            print("Given file already exists, do you want to overwrite it?")
            overwrite = ""
            while True:
                overwrite = input("yes/no: ")
                if overwrite in ["yes", "no"]:
                    break
                else:
                    print("Please give an input of 'yes' or 'no'!")

            if overwrite == "yes":
                self.processor.save_state_to_file(file_path)
                return
            else:
                return

    def print_list_command(self, user_input):
        # no additional commands so only print
        if len(user_input) == 1:
            print("\n")
            for name in self.print_list:
                self.processor.get_module(name).print_module()
                print("\n")
            return

        # add name to print list
        if user_input[1] == "add":
            name = user_input[2]
            # check if name exists
            try:
                self.processor.get_module(name)
            except KeyError:
                print("\nNo module named", name, "\n")
                return

            self.print_list.append(name)

        elif user_input[1] == "remove":
            name = user_input[2]
            if name in self.print_list:
                self.print_list.remove(name)
            else:
                print(name, "not in print list")
        else:
            print("Unknown command")


if __name__ == '__main__':
    cli = CLI(None)
    cli.run()
