from __future__ import annotations


class Module:
    """
    This class specifies the basic functionality of all module types.
    The modules are processor componenets in the processor,
    such as ALU, busses, registers and more.
    """

    def __init__(self, name: str = "") -> None:

        self.name = name

    def update_register(self) -> None:
        """
        Simulates module behaviour for saving input into the internal state
        during a clock tick. Exact behaviour is specifid in each module type.
        """
        pass

    def output_register(self) -> None:
        """
        Simulates module behaviour for giving data to output signals
        during a clock tick. Exact behaviour is specifid in each module type.
        """
        pass

    def update_logic(self) -> None:
        """
        Simulates behaviour for sending output using the internal state
        during a clock tick. Exact behaviour is specifid in each module type.
        """
        pass

    def get_state(self) -> dict:
        """
        Returns a dict of the module states.
        """
        state_dict = dict()
        state_dict["name"] = self.name
        return state_dict

    def get_gui_state(self) -> dict:
        """
        Returns a dict of the module states that should be displayed in a GUI.
        """
        state_dict = dict()
        state_dict["name"] = self.name
        return state_dict

    def set_state(self, state: dict) -> None:
        """
        Sets the modules state to one given in dict. Dict format will be
        diffrent for each type of module so use get_state to get correct format.
        """
        self.name = state["name"]

    def get_output_signals(self) -> []:
        """
        Returns the modules output signals. Is implemented in specific classes.
        """
        return []

    def get_input_signals(self) -> []:
        """
        Returns the modules input signals. Is implemented in specific classes.
        """
        return []

    def reset(self) -> None:
        """
        Resets the module to its default state.
        """
        pass

    def load_from_str(self, state_string):
        """
        Sets the moudles state according to a string
        Each module type will decide how the string should be formated.
        """
        pass

    def get_longest_line_len(self, ignore_keys=[]) -> int:
        """
        Helper function for pretty_print that returns the length of
        the longest line to print for a module.
        """

        longest_line_len = 0
        state = self.get_state()

        for key in state:
            if key not in ignore_keys:
                line = key + ": " + str(state[key])
                if len(line) > longest_line_len:
                    longest_line_len = len(line)

        return longest_line_len

    def print_module(self) -> None:
        """
        Prints the module directly to terminal
        """
        print(self.name)

    def save_state_to_file(self, file_path: str) -> bool:
        """
        Tries to save the modules state to a given file.
        Returns true on success and false if something went wrong.
        """
        pass

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        return f"Module({self.name!r})"
