from __future__ import annotations

from simudator.core.module import Module
from simudator.core.signal import Signal


class Memory(Module):
    """
    A general size memory module.
    """

    def __init__(
        self,
        input_signal: Signal,
        output_signal: Signal,
        control_signal: Signal,
        adress_signal: Signal,
        size: int,
        value_padding: int = 2,
        adress_padding: int = 2,
        name="Memory",
    ) -> None:
        super().__init__(name)

        # Input signals
        self.input_s = input_signal
        self.control_s = control_signal
        self.adress_s = adress_signal

        # Output signals
        self.output_s = output_signal

        # Internal state
        self.memory = [0 for i in range(size)]
        self.current_adress = 0
        self.is_write = False

        # Register self as destination of input signals
        self.input_s.add_destination(self)
        self.control_s.add_destination(self)
        self.adress_s.add_destination(self)

        # Values used to format the strings when saving state to file
        self.adress_padding = adress_padding
        self.value_padding = value_padding

    def update_register(self):
        if self.is_write:
            self.memory[self.current_adress] = self.input_s.get_value()

    def output_register(self) -> None:
        pass
        # Might need to output the current value here
        # in case the memory was written to in update_register()

    def update_logic(self) -> None:
        if (
            self.adress_s.get_value() is not None
            and self.control_s.get_value() is not None
        ):
            self.is_write = self.control_s.get_value()
            self.current_adress = self.adress_s.get_value()
            self.output_s.update_value(self.memory[self.current_adress])

    def get_input_signal(self) -> Signal:
        return self.input_s

    def get_output_signal(self) -> Signal:
        return self.output_s

    def get_adress_signal(self) -> Signal:
        return self.adress_s

    def get_control_signal(self) -> Signal:
        return self.control_s

    def print_module(self) -> None:
        print("", self.name, "\n -----")
        for adress, value in enumerate(self.memory):
            if value:
                print("", str(hex(adress)), "", str(hex(value)), "\n")

    def get_state(self) -> dict:
        state = super().get_state()
        state["is_write"] = self.is_write
        state["current_adress"] = self.current_adress
        state["memory"] = self.memory[:]
        return state

    def get_gui_state(self) -> dict:
        state = super().get_gui_state()
        state["current_adress"] = self.current_adress
        state["memory"] = self.memory[:]
        return state

    def set_state(self, state: dict) -> None:
        super().set_state(state)
        if "is_write" in state:
            self.is_write = state["is_write"]
        if "current_adress" in state:
            self.current_adress = state["current_adress"]
        if "memory" in state:
            self.memory = state["memory"]

    def reset(self) -> None:
        """
        Reset the memory to 0 for each adress.
        """
        for i in range(len(self.memory)):
            self.memory[i] = 0

    def get_longest_line_len(self, ignore_keys=[]) -> int:
        """
        Helper function for pretty_print that returns the length of
        the longest value in the memory to print for a module.
        """

        longest_memory_line = 0

        for value in self.memory:
            value_len = len(str(value))
            if value_len > longest_memory_line:
                longest_memory_line = value_len

        return longest_memory_line

    def get_largest_mem_adr(self) -> int:
        """
        Helper function for pretty_print that returns the length of
        the largest adress in the memory to print for a module.
        """
        return len(str(len(self.memory)))

    def save_state_to_file(self, file_path: str) -> None:
        """
        Tries to save the modules state to a given file.
        """
        file = open(file_path, "a")
        file.write(self.name + ":\n")
        for index, value in enumerate(self.memory):
            # Write the adress in hex and add ': ' to the end
            file.write(hex(index)[2:].rjust(self.adress_padding, "0") + ": ")
            # Write the value in hex
            file.write(hex(value)[2:].rjust(self.value_padding, "0"))
            file.write("\n")

        file.write("\n")
        file.close()
