from __future__ import annotations

from typing import Any

from simudator.core.module import Module
from simudator.core.signal import Signal


class Register(Module):
    """
    A simple module that can store a value.
    """

    def __init__(self, input_signal: Signal,
                       output_signal: Signal,
                       value: Any = 0, name: str = "Register") -> None:

        # set the registers name
        super().__init__(name)

        # signals
        self.input_s = input_signal
        self.output_s = output_signal
        self.value = value

        # set the register as destination of its input signal
        self.input_s.add_destination(self)

    def update_register(self) -> None:
        """
        Propagate the input signal to the registers internal state.
        Throw away bits larger than the length of the register.
        """
        input_value = self.input_s.get_value()
        if input_value is None:
            return
        self.value = input_value

    def output_register(self) -> None:
        """
        Propagate the value of the register to the output signal.
        It is the response of the destination to know when it should
        read the output.
        """
        self.output_s.update_value(self.value)

    def update_logic(self):
        """
        The register has no logic.
        """
        pass

    def get_state(self) -> dict[str: Any]:
        """
        Returns a dict of the register state.
        These states are changable via set_states.
        """
        state = super().get_state()
        state["value"] = self.value
        return state

    def get_gui_state(self) -> dict:
        state = self.get_state()
        return state

    def set_state(self, state: dict[str: Any]) -> None:
        """
        Sets the register state to one given in dict.
        """
        self.name = state["name"]
        self.value = state["value"]

    def reset(self) -> None:
        """
        Resets the register to 0.
        """
        self.value = 0

    def get_output_signals(self) -> []:
        return [self.output_s]

    def get_input_signals(self) -> []:
        return [self.input_s]

    def save_state_to_file(self, file_path: str) -> None:
        """
        Tries to save the modules state to a given file.
        Returns true on success and false if something went wrong.
        """
        file = open(file_path, "a")
        file.write(self.name + ":\n")
        file.write("value: " + str(self.value) + "\n\n")
        file.close()

    def print_module(self) -> None:
        print("", self.name, "\n -----",
              "\n value: ", self.value,
              )


class IntegerRegister(Register):
    """
    A simple module that can store an integer value with a given bit-length.
    """

    def __init__(self, input_signal: Signal, output_signal: Signal,
                 bit_length: int, value: int = 0, name: str | None = None) -> None:

        # set the registers name
        if name is None:
            name = f"{bit_length}-bit register"

        super().__init__(input_signal, output_signal, value=value, name=name)

        # set the bit length of the register
        self.bit_length = bit_length

        # set the registers mask. An 8 bit register should
        # have the mask 1111 1111, aka one '1' for every bit
        self.mask = 2**bit_length - 1

    def update_register(self) -> None:
        """
        Propagate the input signal to the registers internal state.
        Throw away bits larger than the length of the register.
        """
        super().update_register()
        self.value = self.value & self.mask

    def get_state(self) -> dict[str: Any]:
        """
        Returns a dict of the register state.
        These states are changable via set_states.
        """
        state = super().get_state()
        state["bit_length"] = self.bit_length
        return state

    def set_state(self, state: dict[str: Any]) -> None:
        """
        Sets the register state to one given in dict.
        """
        super().set_state(state)
        if "bit_length" in state:
            self.bit_length = state["bit_length"]
            self.mask = 2**self.bit_length -1

    def save_state_to_file(self, file_path: str) -> None:
        """
        Tries to save the modules state to a given file.
        """
        file = open(file_path, "a")
        file.write(self.name + ":\n")
        file.write("value: " + hex(self.value)[2:] + "\n\n")
        file.close()

    def load_from_str(self, state_string):
        string_pair = state_string.split(": ")
        #TODO: Maybe check if it starts with value: ?
        self.value = int(string_pair[1], 16)


class Flag(IntegerRegister):
    """
    A simple module that can store a value. This module is also
    updated during the 'update logic' phase.
    """

    def __init__(self, input_signal: Signal, output_signal: Signal,
                 bit_length=1, value=0, name="Flag") -> None:

        # set the flags name
        super().__init__(input_signal=input_signal, output_signal=output_signal,
                         bit_length=bit_length, value=value, name=name)

    def output_register(self) -> None:
        """
        Propagate the value of the flag to the output signal.

        It is the response of the destination to know when it should
        read the output.
        """
        self.output_s.update_value(self.value)

    def update_logic(self):
        """
        The flag has no logic but it still need to be updated during
        the update logic phase. This is due the fact that the flag
        should update the same tick as the change that causes the flag
        to activate (ex: loop counter reaches zero -> set L flag to 1)
        """
        input_value = self.input_s.get_value()
        self.value = input_value & self.mask

    def get_gui_state(self) -> dict:
        state = super().get_gui_state()
        state["value"] = self.value
        return state

    def save_state_to_file(self, file_path: str) -> None:
        """
        Tries to save the modules state to a given file.
        """
        file = open(file_path, "a")
        file.write(self.name + ":\n")
        file.write("value: " + str(self.value) + "\n\n")
        file.close()

    def print_module(self) -> None:
        print("", self.name, "\n -----",
              "\n value: ", self.value,
              "\n bit length: ", self.bit_length,
              "\n mask: ", self.mask,
              )
