from qtpy.QtCore import Qt
from qtpy.QtCore import Signal as pyqtSignal
from qtpy.QtCore import Slot
from qtpy.QtWidgets import (
    QAction,
    QGraphicsRectItem,
    QGraphicsSimpleTextItem,
    QTextEdit,
    QVBoxLayout,
    QWidget,
)

from simudator.core.modules import Memory
from simudator.gui.dialogs.memory_content_dialog import MemoryContentDialog
from simudator.gui.module_graphics_item.module_graphics_item import ModuleGraphicsItem
from simudator.gui.orientation import Orientation
from simudator.gui.port_graphics_item import PortGraphicsItem


class MemoryWindow(QWidget):
    """
    Widget for showing content of memory
    """

    _ROW_LENGTH = 5

    def __init__(self, memory_module: Memory):
        super().__init__()
        self.module = memory_module

        self.text = QTextEdit("")
        layout = QVBoxLayout()
        layout.addWidget(self.text)
        self.setLayout(layout)

        self.update()

    def update(self):
        memory_str = ""
        for adress, value in enumerate(self.module.memory):
            # Add adress and content to string
            # Make sure its unifrom lenght so rows are consistent
            memory_str += f"{adress}" + ": " + f"{value}" + "   " + "\t"

            # Make new line when we reach end of row
            if adress % self._ROW_LENGTH == self._ROW_LENGTH - 1:
                memory_str += "\n"

        self.text.setText(memory_str)


class MemoryGraphicsItem(ModuleGraphicsItem):
    """
    Graphics module for a Memory module.
    """

    new_memory_breakpoint_signal = pyqtSignal(str, str, str)

    def __init__(self, memory_module: Memory, width: int = 150, height: int = 200):
        self._width = width
        self._height = height
        super().__init__(memory_module)
        self.memory_window = None

    def draw_graphics_item(self) -> None:
        # Create base backgrond rect
        self.baserect = QGraphicsRectItem(0, 0, self._width, self._height, self)

        # Make ports
        self.input = PortGraphicsItem(
            self.module.get_input_signal(), Orientation.RIGHT, self
        )
        self.input.setPos(self._width, self._height / 8)

        self.output = PortGraphicsItem(
            self.module.get_output_signal(), Orientation.RIGHT, self
        )
        self.output.setPos(self._width, self._height / 4)

        self.adress = PortGraphicsItem(
            self.module.get_adress_signal(), Orientation.DOWN, self
        )
        self.adress.setPos(self._width / 4, self._height)

        self.control = PortGraphicsItem(
            self.module.get_control_signal(), Orientation.DOWN, self
        )
        self.control.setPos(self._width * 3 / 4, self._height)

        self.ports = [self.input, self.output, self.adress, self.control]

        # Create name lable
        name_text = QGraphicsSimpleTextItem(self.state["name"], self)
        name_text.setPos(self._width / 2 - len(self.state["name"]) * self._CHAR_LEN, 0)

        # Create adress lable
        self.adress_text = QGraphicsSimpleTextItem(
            "adress: " + str(self.state["current_adress"]), self
        )
        self.adress_text.setPos(self._width / 20, self._height / 8)

    def update(self):
        """
        Also update memory window here.
        """
        self.state = self.module.get_state()
        self.adress_text.setText("adress: " + str(self.state["current_adress"]))
        if self.memory_window is not None:
            self.memory_window.update()

    def mousePressEvent(self, event):
        """
        Show content window if module is right clicked
        """
        return
        # TODO: comment =/= what fuction does
        if event.button() == Qt.MouseButton.RightButton:
            self.memory_window = MemoryWindow(self.module)
            self.memory_window.show()
            super().mousePressEvent(event)

    def showMemoryContents(self) -> None:
        """
        Create and show a MemoryWindow that displays the contents
        of the memory module associated with this graphics item.
        """
        self.memory_window = MemoryWindow(self.module)
        self.memory_window.show()

    def shouldIgnoreAction(self, action: QAction) -> bool:
        """
        Returns True if the given action should not be used in the context menu
        of this graphics item.
        """
        if action.data() == "MEM_ONLY":
            return False
        return super().shouldIgnoreAction(action)

    def generateActions(self) -> []:
        # Make parent actions and then add memory specific actions
        super().generateActions()

        memory_br_action = QAction('Add memory breakpoint', self)
        memory_br_action.triggered.connect(self.memoryBreakpointDialog)
        self.actions.append(memory_br_action)

        edit_memory_action = QAction('Edit memory content', self)
        edit_memory_action.triggered.connect(self.editMemoryContentDialog)
        self.actions.append(edit_memory_action)

        show_content_action = QAction('Show memory content', self)
        show_content_action.triggered.connect(self.showMemoryContents)
        self.actions.append(show_content_action)

    def memoryBreakpointDialog(self) -> None:
        """
        Opens dialog window for user to create a breakpoint.
        """
        self.memory_br_dialog = MemoryContentDialog(self.module)
        self.memory_br_dialog.accepted.connect(self.memoryBreakpointAccepted)

    def editMemoryContentDialog(self) -> None:
        """
        Opens dialog window for user to edit memory.
        """
        self.memory_edit_dialog = MemoryContentDialog(self.module)
        self.memory_edit_dialog.accepted.connect(self.editMemoryAccepted)

    @Slot(str, str, str)
    def memoryBreakpointAccepted(
        self, module_name: str, adress: str, value: str
    ) -> None:
        """
        Takes the info from a breakpoint dialog and sends it to the gui
        """
        try:
            self.new_memory_breakpoint_signal.emit(module_name, adress, value)
        except SyntaxError as e:
            self.errorMessageWidget.showMessage(str(e))

    @Slot(str, str, str)
    def editMemoryAccepted(self, module_name: str, adress: str, value: str) -> None:
        """
        Takes the info from a edit memory dialog
        and updates the memory content accordingly.
        Also asks the gui to update.
        """
        try:
            parsed_adress = int(adress)
        except SyntaxError as e:
            self.errorMessageWidget.showMessage(str(e))
        else:
            module_state = self.module.get_state()
            module_state['memory'][parsed_adress] = value
            self.module.set_state(module_state)
            self.update_graphics_signal.emit()

    def getActionSignals(self) -> []:
        # Do parent signals and then add memory specific signals
        signals = super().getActionSignals()
        signals.append(("MEM_BP", self.new_memory_breakpoint_signal))
        return signals
