from __future__ import annotations

from typing import Any

from simudator.core.module import Module
from simudator.core.signal import Signal
from simudator.processor.mia.modules.mia_bus_connect import MiaBusConnector


class GRX(Module, MiaBusConnector):
    """
    A class representing the four registers and the mux/demux
    used in the mia architecture. The S signal determines if the
    registers should be indexed by the GRx bits or the M bits.
    """

    def __init__(
        self,
        to_bus: Signal,
        from_bus: Signal,
        bus_control: Signal,
        grx_control: Signal,
        m_control: Signal,
        s_control: Signal,
        bus_id: int,
        name="GRx",
        registers=[0 for _ in range(4)],
        bit_length=16,
    ) -> None:
        # Set connection to/from bus and bus_id
        MiaBusConnector.__init__(self, bus_control, bus_id)

        # Set the name
        Module.__init__(self, name)

        # Other signals
        self.to_bus = to_bus
        self.from_bus = from_bus
        self.m_control_s = m_control
        self.s_control_s = s_control
        self.grx_control_s = grx_control

        # Set self as destination of its signals
        self.m_control_s.add_destination(self)
        self.s_control_s.add_destination(self)
        self.grx_control_s.add_destination(self)

        # set the regusters
        self.registers = registers

        # set the bit_length
        self.bit_length = bit_length

    def update_register(self):
        """
        If the module should read from the bus, update the correct
        register by reading the s bit and index the registers by
        the GRx bits if s == 0 or M bits if s == 1.
        """
        if self.read_from_bus():
            s_control_value = self.s_control_s.get_value()

            grx_index = 0
            if s_control_value == 0b0:
                grx_index = self.grx_control_s.get_value()
            elif s_control_value == 0b1:
                grx_index = self.m_control_s.get_value()
            else:
                raise ValueError("The value of the s signal" "must be 0b0 or 0b1 !!!")

            self.registers[grx_index] = self.from_bus.get_value()

    def output_register(self):
        """
        If the module should write to the bus, update the bus
        by reading the s bit and index the registers by
        the GRx bits if s == 0 or M bits if s == 1.
        """
        if self.write_to_bus():
            s_control_value = self.s_control_s.get_value()

            grx_index = 0
            if s_control_value == 0b0:
                grx_index = self.grx_control_s.get_value()
            elif s_control_value == 0b1:
                grx_index = self.m_control_s.get_value()
            else:
                raise ValueError("The value of the s signal" "must be 0b0 or 0b1 !!!")

            self.to_bus.update_value(self.registers[grx_index])
        else:
            self.to_bus.update_value(None)

    def update_logic(self) -> None:
        self.output_register()

    def get_state(self) -> dict[str, Any]:
        """
        Returns a dict of the grx state.
        These states are changable via set_states.
        """
        state = {
            "name": self.name,
            "bit_length": self.bit_length,
            "registers": self.registers[:],
        }
        return state

    def get_gui_state(self) -> dict:
        state = {
            "name": self.name,
            "registers": self.registers[:],
        }
        return state

    def set_state(self, state: dict[str, Any]):
        """
        Sets the grx state to one given in dict.
        """
        self.name = state["name"]
        if "bit_length" in state:
            self.bit_length = state["bit_length"]
        self.registers = state["registers"]

    def reset(self) -> None:
        """
        Resets the GRx registers to 0.
        """
        for i in range(len(self.registers)):
            self.registers[i] = 0

    def save_state_to_file(self, file_path: str) -> None:
        """
        Tries to save the modules state to a given file.
        Returns true on success and false if something went wrong.
        """
        file = open(file_path, "a")
        file.write(self.name + ":\n")
        for index, value in enumerate(self.registers):
            file.write(hex(index)[2:] + ": ")
            file.write(hex(value)[2:])
            file.write("\n")

        file.write("\n")
        file.close()

    def load_from_str(self, state_string) -> None:
        """
        Loads the module from a string, used when loading state from
        files.
        """
        self.reset()
        lines = state_string.split("\n")
        for line in lines:
            if line:  # last line is empty from split with \n
                line_data = line.split(": ")
                adress = int(line_data[0], 16)
                value = int(line_data[1], 16)
                self.registers[adress] = value

    def print_module(self):
        print(
            "",
            self.name,
            "\n -----",
            "\n bus_id: ",
            self.bus_id,
            "\n bit_length: ",
            self.bit_length,
            "\n GR0: ",
            hex(self.registers[0]),
            "\n GR1: ",
            hex(self.registers[1]),
            "\n GR2: ",
            hex(self.registers[2]),
            "\n GR3: ",
            hex(self.registers[3]),
        )

    def get_input_signals(self) -> list[Signal]:
        return [
            self.bus_control_s,
            self.from_bus,
            self.m_control_s,
            self.s_control_s,
            self.grx_control_s,
        ]

    def get_output_signals(self) -> list[Signal]:
        return [self.to_bus]
