from simudator.core.module import Module
from simudator.core.signal import Signal


class MicroPC(Module):
    """
    A class representing the micro program counter (uPC) of the MIA-system.
    The module reads signal data from the memory modules K1 and K2, the
    register SuPC and the micro memory uM. The module writes signal
    data to SuPC and uM. A control signal is used to determine which signal
    to read from or write to.
    """

    def __init__(
        self,
        control_signal: Signal,
        from_k1: Signal,
        from_k2: Signal,
        to_supc: Signal,
        from_supc: Signal,
        to_um: Signal,
        from_um: Signal,
        bit_length: int = 8,
        name: str = "uPC",
    ) -> None:
        super().__init__(name)

        self.value = 0
        self.bit_length = bit_length

        # Input signals
        self.control_signal = control_signal
        self.from_k1 = from_k1
        self.from_k2 = from_k2
        self.from_supc = from_supc
        self.from_um = from_um

        # Output signals
        self.to_supc = to_supc
        self.to_um = to_um

        # Register as destination of the input signals
        control_signal.add_destination(self)
        from_k1.add_destination(self)
        from_k2.add_destination(self)
        from_supc.add_destination(self)
        from_um.add_destination(self)

    def update_register(self) -> None:
        match self.control_signal.get_value():
            case 0b000:
                self.value += 1

                # Overflow
                if self.value == 128:
                    self.value = 0

            case 0b001:
                self.value = self.from_k1.get_value()
            case 0b010:
                self.value = self.from_k2.get_value()
            case 0b011:
                self.value = 0
            case 0b100:
                self.value = self.from_um.get_value()
            case 0b101:
                self.value = self.from_supc.get_value()
            case 0b110:
                self.to_supc.update_value(self.value + 1)
                self.value = self.from_um.get_value()

    def output_register(self) -> None:
        self.to_um.update_value(self.value)

    def get_state(self) -> dict:
        state = super().get_state()
        state["value"] = self.value
        state["bit_length"] = self.bit_length
        return state

    def get_gui_state(self) -> dict:
        state = {
            "name": self.name,
            "value": self.value,
        }
        return state

    def set_state(self, state: dict) -> None:
        super().set_state(state)
        self.value = state["value"]
        if "bit_length" in state:
            self.bit_length = state["bit_length"]

    def reset(self) -> None:
        """
        Resets the micro program counter to 0.
        """
        self.value = 0

    def save_state_to_file(self, file_path: str) -> None:
        """
        Tries to save the modules state to a given file.
        """
        file = open(file_path, "a")
        file.write(self.name + ":\n")
        file.write("value: " + hex(self.value)[2:] + "\n\n")
        file.close()

    def load_from_str(self, state_string):
        string_pair = state_string.split(": ")
        # TODO: Maybe check if it starts with value: ?
        self.value = int(string_pair[1], 16)

    def print_module(self) -> None:
        print("", self.name, "\n -----", "\n value: ", hex(self.value))

    def get_input_signals(self) -> list[Signal]:
        return [
            self.control_signal,
            self.from_k1,
            self.from_k2,
            self.from_supc,
            self.from_um,
        ]

    def get_output_signals(self) -> list[Signal]:
        return [
            self.to_supc,
            self.to_um,
        ]
