from __future__ import annotations

from typing import Any

from simudator.core.module import Module
from simudator.core.signal import Signal
from simudator.processor.mia.modules.mia_bus_connect import MiaBusConnector


class PC(Module, MiaBusConnector):
    """A program counter moduled after the mia lab.

    The module have two
    control signals, p and bus_control. If the p signal is true the
    program counter increases its value by one, if the bus_control
    is true (the id to read from bus is the pc's id) the pc takes the
    value on the bus. If both are true it does nothing.
    """

    def __init__(
        self,
        p: Signal,
        bus_input: Signal,
        bus_output: Signal,
        bus_control: Signal,
        bus_id: int,
        name="PC",
        value=0,
    ) -> None:
        MiaBusConnector.__init__(self, bus_control, bus_id)
        Module.__init__(self, name)

        # control signals
        self.p = p

        # input/output signals
        self.bus_input_s = bus_input
        self.bus_output_s = bus_output

        # set program counter as destination of all signlas
        self.p.add_destination(self)
        self.bus_input_s.add_destination(self)

        # internal state
        self.value = value

        # help variables
        self.increase_by_one = False

    def output_register(self) -> None:
        """Output the value of the program counter to the bus."""
        if self.write_to_bus():
            self.bus_output_s.update_value(self.value)
        else:
            self.bus_output_s.update_value(None)

    def update_register(self) -> None:
        """Updates the value of the PC according to the signals sent
        by the micro memory. If the program counter recives
        'True' from both signals it should do nothing and wait for the
        next cycle.
        """

        self.increase_by_one = self.p.get_value()

        if self.increase_by_one and self.read_from_bus():
            return

        if self.read_from_bus():
            # Read the 8 lowest bits from the bus
            self.value = self.bus_input_s.get_value() & 0xFF

        if self.increase_by_one:
            self.value += 1

    def update_logic(self) -> None:
        self.output_register()

    def get_state(self) -> dict[str:Any]:
        """Returns a dict of the program counter state.
        These states are changable via set_states.

        Returns
        -------
        dict[Any]
            The state of the program counter.
        """
        state = dict()
        state["name"] = self.name
        state["value"] = self.value
        state["increment"] = self.increase_by_one
        state["bit_length"] = 8
        return state

    def get_gui_state(self) -> dict:
        state = {
            "name": self.name,
            "value": self.value,
        }
        return state

    def set_state(self, state: dict[str:Any]) -> None:
        """Sets the program counter state to one given in dict."""
        self.name = state["name"]
        self.value = state["value"]
        if "increment" in state:
            self.increase_by_one = state["increment"]

    def reset(self) -> None:
        """Reset the program counter to 0."""
        self.value = 0
        self.increase_by_one = False

    def save_state_to_file(self, file_path: str) -> None:
        """Tries to save the modules state to a given file."""
        file = open(file_path, "a")
        file.write(self.name + ":\n")
        file.write("value: " + hex(self.value)[2:] + "\n\n")
        file.close()

    def load_from_str(self, state_string):
        string_pair = state_string.split(": ")
        # TODO: Maybe check if it starts with value: ?
        self.value = int(string_pair[1], 16)

    def print_module(self) -> None:
        print("", self.name, "\n -----", "\n value: ", hex(self.value))

    def get_input_signals(self) -> list[Signal]:
        return [self.bus_input_s, self.p]

    def get_output_signals(self) -> list[Signal]:
        return [self.bus_output_s]
