from simudator.core.processor import Processor
from simudator.core.signal import Signal
from simudator.processor.mia.modules.mia_memory import MiaMemory


def test_read():
    cpu = Processor()
    input_s = Signal(cpu, value=0)
    output_s = Signal(cpu, value=0)
    adress_s = Signal(cpu, value=0)
    bus_control_s = Signal(cpu, value=0)
    mem = MiaMemory(input_s, output_s, adress_s, bus_control_s, 4, 0b010)
    cpu.add_module(mem)

    # Set up memory content
    state = mem.get_state()
    state["memory"] = [1, 2, 3, 4]
    mem.set_state(state)

    # Set bus control signal to read from memory but not write to it
    bus_control_s.update_value((0b010, 0))

    # Test reading from first adress, should be 1
    cpu.do_tick()
    assert output_s.get_value() == 1

    # Test reading from last adress, should be 4
    adress_s.update_value(3)
    mem.update_logic()
    cpu.do_tick()
    assert output_s.get_value() == 4

    # Test that the memory outputs None when not reading from the memory
    bus_control_s.update_value((0, 0))
    mem.update_logic()
    assert output_s.get_value() is None
    cpu.do_tick()
    assert output_s.get_value() is None


def test_write():
    cpu = Processor()
    input_s = Signal(cpu, value=0)
    output_s = Signal(cpu, value=0)
    adress_s = Signal(cpu, value=0)
    bus_control_s = Signal(cpu, value=0)
    mem = MiaMemory(input_s, output_s, adress_s, bus_control_s, 4, 0b010)
    cpu.add_module(mem)

    # Set up memory content
    init_state = mem.get_state()
    init_state["memory"] = [1, 2, 3, 4]
    mem.set_state(init_state)

    # Set bus control signal to write to and read from the memory
    bus_control_s.update_value((0b010, 0b010))

    # Set up for writing to the memory
    input_s.update_value(5)

    # Check that nothing is written to the memory until the clock has signalled
    mem.update_logic()
    assert mem.get_state()["memory"] == init_state["memory"]

    # Write to and then read from the first adress
    cpu.do_tick()
    assert output_s.get_value() == 5
    assert mem.get_state()["memory"] == [5, 2, 3, 4]

    # Test writing to but not reading from the memory
    bus_control_s.update_value((0, 0b010))
    adress_s.update_value(2)
    input_s.update_value(6)
    mem.update_logic()
    cpu.do_tick()
    assert output_s.get_value() is None
    assert mem.get_state()["memory"] == [5, 2, 6, 4]


def test_load():
    PATH = "test/test_mia/mia_memory_data_test.txt"
    cpu = Processor()
    input_s1 = Signal(cpu, value=0)
    output_s1 = Signal(cpu, value=0)
    adress_s1 = Signal(cpu, value=0)
    bus_control_s1 = Signal(cpu, value=0)
    input_s2 = Signal(cpu, value=0)
    output_s2 = Signal(cpu, value=0)
    adress_s2 = Signal(cpu, value=0)
    bus_control_s2 = Signal(cpu, value=0)
    pm1 = MiaMemory(
        input_s1, output_s1, adress_s1, bus_control_s1, 4, 0b010, name="PM1"
    )
    pm2 = MiaMemory(
        input_s2, output_s2, adress_s2, bus_control_s2, 4, 0b010, name="PM2"
    )
    cpu.add_module(pm1)
    cpu.add_module(pm2)
    cpu.load_state_from_file(PATH)

    adress_s1.update_value(0)
    adress_s2.update_value(0)
    bus_control_s1.update_value((0b010, 0))
    bus_control_s2.update_value((0b010, 0))
    cpu.do_tick()

    assert output_s1.get_value() == 0x0A
    assert output_s2.get_value() == 0

    adress_s1.update_value(1)
    adress_s2.update_value(1)
    bus_control_s1.update_value((0b010, 0))
    bus_control_s2.update_value((0b010, 0))
    cpu.do_tick()

    assert output_s1.get_value() == 0
    assert output_s2.get_value() == 0x02

    adress_s1.update_value(2)
    adress_s2.update_value(2)
    bus_control_s1.update_value((0b010, 0))
    bus_control_s2.update_value((0b010, 0))
    cpu.do_tick()

    assert output_s1.get_value() == 0x0F
    assert output_s2.get_value() == 0

    adress_s1.update_value(3)
    adress_s2.update_value(3)
    bus_control_s1.update_value((0b010, 0))
    bus_control_s2.update_value((0b010, 0))
    cpu.do_tick()

    assert output_s1.get_value() == 0
    assert output_s2.get_value() == 0
